use tokio::sync::{mpsc, oneshot};

use super::registry::Registry;

pub trait BrokerRequestInternal: Send {
    fn run_request(self: Box<Self>, registry: &mut Registry);
}

pub trait MakeBrokerRequest: BrokerRequestInternal {
    type Response;

    fn create(response_tx: oneshot::Sender<Self::Response>) -> Box<Self>;

    fn new() -> (Box<Self>, oneshot::Receiver<Self::Response>) {
        let (response_tx, response_rx) = oneshot::channel();
        (Self::create(response_tx), response_rx)
    }
}

pub trait RequestHandler: Send {
    type Response: Send;

    fn handle(self, registry: &mut Registry) -> Self::Response;
}

pub struct BrokerRequest<H: RequestHandler> {
    handler: H,
    response_tx: oneshot::Sender<H::Response>,
}

impl<H: RequestHandler> BrokerRequestInternal for BrokerRequest<H> {
    fn run_request(self: Box<Self>, registry: &mut Registry) {
        let response = self.handler.handle(registry);
        let _ = self.response_tx.send(response);
    }
}

impl<H: RequestHandler + Default> MakeBrokerRequest for BrokerRequest<H> {
    type Response = H::Response;

    fn create(response_tx: oneshot::Sender<H::Response>) -> Box<Self> {
        Box::new(Self {
            handler: H::default(),
            response_tx,
        })
    }
}

pub type BrokerRequestBox = Box<dyn BrokerRequestInternal>;
pub type BrokerRequestSender = mpsc::Sender<BrokerRequestBox>;
